import uuid
from copy import copy, deepcopy
from datetime import datetime
from typing import TYPE_CHECKING, Set

import attr
from six.moves.urllib.parse import urlparse, urlunsplit

from applitools.common import deprecated
from applitools.common.accessibility import AccessibilitySettings
from applitools.common.geometry import RectangleSize
from applitools.common.match import ImageMatchSettings, MatchLevel
from applitools.common.server import FailureReports, SessionType
from applitools.common.utils import argument_guard
from applitools.common.utils.converters import str2bool
from applitools.common.utils.datetime_utils import UTC
from applitools.common.utils.general_utils import get_env_with_prefix
from applitools.common.utils.json_utils import JsonInclude

if TYPE_CHECKING:
    from typing import TYPE_CHECKING, Dict, List, Optional, Text, TypeVar

    from applitools.common.utils.custom_types import ViewPort

    Self = TypeVar("Self", bound="Configuration")  # typedef

__all__ = ("BatchInfo", "Configuration")

MINIMUM_MATCH_TIMEOUT_MS = 600
DEFAULT_MATCH_TIMEOUT_MS = 2000  # type: int
DEFAULT_SERVER_REQUEST_TIMEOUT_MS = 60 * 5 * 1000
DEFAULT_ALL_TEST_RESULTS_TIMEOUT = 30 * 60  # seconds
DEFAULT_SERVER_URL = "https://eyesapi.applitools.com"
PROCESS_DEFAULT_BATCH_ID = str(uuid.uuid4())  # Unique per-process


@attr.s(init=False, slots=True)
class BatchInfo(object):
    """
    A batch of tests.
    """

    name = attr.ib(metadata={JsonInclude.THIS: True})  # type: Text
    started_at = attr.ib(metadata={JsonInclude.THIS: True})  # type: datetime
    sequence_name = attr.ib(
        metadata={JsonInclude.NAME: "batchSequenceName"}
    )  # type: Optional[Text]
    id = attr.ib(metadata={JsonInclude.THIS: True})  # type: Text
    notify_on_completion = attr.ib(metadata={JsonInclude.NON_NONE: True})  # type: bool
    properties = attr.ib(
        metadata={JsonInclude.NON_NONE: True}
    )  # type: List[Dict[Text,Text]]

    def __init__(self, name=None, started_at=None, batch_sequence_name=None):
        # type: (Optional[Text], Optional[datetime], Optional[Text]) -> None
        self.id = get_env_with_prefix(
            "APPLITOOLS_BATCH_ID", PROCESS_DEFAULT_BATCH_ID
        )  # type: Text
        self.name = get_env_with_prefix("APPLITOOLS_BATCH_NAME")  # type: Text
        self.started_at = datetime.now(UTC)  # type: datetime
        self.sequence_name = get_env_with_prefix(
            "APPLITOOLS_BATCH_SEQUENCE"
        )  # type: Optional[Text]
        self.notify_on_completion = str2bool(
            get_env_with_prefix("APPLITOOLS_BATCH_NOTIFY")
        )  # type: bool
        self.properties = []  # type: List[Dict[Text,Text]]

        if name:
            self.name = name
        if started_at:
            self.started_at = started_at
        if batch_sequence_name:
            self.sequence_name = batch_sequence_name

    def with_batch_id(self, id):
        # type: (Text) -> BatchInfo
        argument_guard.not_none(id)
        self.id = str(id)
        return self

    def add_property(self, name, value):
        # type: (Text, Text) -> BatchInfo
        """
        Associates a key/value pair with the Batch. This can be used later for filtering.
        :param name: (string) The property name.
        :param value: (string) The property value
        """
        self.properties.append({"name": name, "value": value})
        return self

    def clear_properties(self):
        # type: () -> BatchInfo
        """
        Clears the list of Batch properties.
        """
        del self.properties[:]
        return self


class ProxySettings(object):
    def __init__(
        self,
        host_or_url,  # type: Text
        port=None,  # type: Optional[int]
        username=None,  # type: Optional[Text]
        password=None,  # type: Optional[Text]
        scheme=None,  # type: Text
    ):
        # type: (...) -> None
        if host_or_url.startswith("http://") or host_or_url.startswith("https://"):
            parsed = urlparse(host_or_url)
            self.host = parsed.hostname
            self.port = port or parsed.port or 8888
            self.username = username or parsed.username
            self.password = password or parsed.password
            self.scheme = scheme or parsed.scheme or "http"
        else:
            self.host = host_or_url
            self.port = port or 8888
            self.username = username
            self.password = password
            self.scheme = scheme or "http"

    @property
    def url(self):
        password = ":{}".format(self.password) if self.password else ""
        auth = "{}{}@".format(self.username, password) if self.username else ""
        port = ":{}".format(self.port) if self.port else ""
        return urlunsplit((self.scheme, auth + self.host + port, "", "", ""))


@attr.s
class Configuration(object):
    batch = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, factory=BatchInfo
    )  # type: BatchInfo
    branch_name = attr.ib(
        metadata={JsonInclude.NON_NONE: True},
        factory=lambda: get_env_with_prefix("APPLITOOLS_BRANCH", None),
    )  # type: Optional[Text]
    parent_branch_name = attr.ib(
        metadata={JsonInclude.NON_NONE: True},
        factory=lambda: get_env_with_prefix("APPLITOOLS_PARENT_BRANCH", None),
    )  # type: Optional[Text]
    baseline_branch_name = attr.ib(
        metadata={JsonInclude.NON_NONE: True},
        factory=lambda: get_env_with_prefix("APPLITOOLS_BASELINE_BRANCH", None),
    )  # type: Optional[Text]
    agent_id = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, default=None
    )  # type: Optional[Text]
    baseline_env_name = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, default=None
    )  # type: Optional[Text]
    environment_name = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, default=None
    )  # type: Optional[Text]
    save_diffs = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, default=None
    )  # type: bool
    app_name = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, default=None
    )  # type: Optional[Text]
    test_name = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, default=None
    )  # type: Optional[Text]
    viewport_size = attr.ib(
        metadata={JsonInclude.NON_NONE: True},
        default=None,
        converter=attr.converters.optional(RectangleSize.from_),
    )  # type: Optional[RectangleSize]
    session_type = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, default=SessionType.SEQUENTIAL
    )  # type: SessionType
    host_app = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, default=None
    )  # type: Optional[Text]
    host_os = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, default=None
    )  # type: Optional[Text]
    properties = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, factory=list
    )  # type: List[Dict[Text, Text]]
    match_timeout = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, default=DEFAULT_MATCH_TIMEOUT_MS
    )  # type: int # ms
    is_disabled = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, default=False
    )  # type: bool
    save_new_tests = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, default=True
    )  # type: bool
    save_failed_tests = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, default=False
    )  # type: bool
    failure_reports = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, default=FailureReports.ON_CLOSE
    )  # type: FailureReports
    send_dom = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, default=True
    )  # type: bool
    default_match_settings = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, factory=ImageMatchSettings
    )  # type: ImageMatchSettings
    stitch_overlap = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, default=None
    )  # type: Optional[int]
    api_key = attr.ib(
        factory=lambda: get_env_with_prefix("APPLITOOLS_API_KEY", None)
    )  # type: Optional[Text]
    server_url = attr.ib(
        metadata={JsonInclude.NON_NONE: True},
        factory=lambda: get_env_with_prefix(
            "APPLITOOLS_SERVER_URL", DEFAULT_SERVER_URL
        ),
    )  # type: Text
    _timeout = attr.ib(default=DEFAULT_SERVER_REQUEST_TIMEOUT_MS)  # type: int # ms
    proxy = attr.ib(default=None)  # type: ProxySettings
    save_debug_screenshots = attr.ib(default=False)  # type: bool
    debug_screenshots_path = attr.ib(
        converter=str, factory=lambda: get_env_with_prefix("DEBUG_SCREENSHOT_PATH", "")
    )
    debug_screenshots_prefix = attr.ib(
        converter=str,
        factory=lambda: get_env_with_prefix("DEBUG_SCREENSHOT_PREFIX", "screenshot_"),
    )

    @property
    def enable_patterns(self):
        # type: () -> bool
        return self.default_match_settings.enable_patterns

    @enable_patterns.setter
    def enable_patterns(self, enable_patterns):
        # type: (bool) -> None
        self.default_match_settings.enable_patterns = enable_patterns

    @property
    def use_dom(self):
        # type: () -> bool
        return self.default_match_settings.use_dom

    @use_dom.setter
    def use_dom(self, use_dom):
        # type: (bool) -> None
        self.default_match_settings.use_dom = use_dom

    @property
    def match_level(self):
        # type: () -> MatchLevel
        return self.default_match_settings.match_level

    @match_level.setter
    def match_level(self, match_level):
        # type: (MatchLevel) -> None
        self.default_match_settings.match_level = match_level

    @property
    def ignore_displacements(self):
        # type: () -> bool
        return self.default_match_settings.ignore_displacements

    @ignore_displacements.setter
    def ignore_displacements(self, ignore_displacements):
        # type: (bool) -> None
        self.default_match_settings.ignore_displacements = ignore_displacements

    def set_batch(self, batch):
        # type: (Self, BatchInfo) -> Self
        argument_guard.is_a(batch, BatchInfo)
        self.batch = batch
        return self

    def set_branch_name(self, branch_name):
        # type: (Self, Text) -> Self
        self.branch_name = branch_name
        return self

    def set_agent_id(self, agent_id):
        # type: (Self, Text) -> Self
        self.agent_id = agent_id
        return self

    def set_parent_branch_name(self, parent_branch_name):
        # type: (Self, Text) -> Self
        self.parent_branch_name = parent_branch_name
        return self

    def set_baseline_branch_name(self, baseline_branch_name):
        # type: (Self, Text) -> Self
        self.baseline_branch_name = baseline_branch_name
        return self

    def set_baseline_env_name(self, baseline_env_name):
        # type: (Self, Text) -> Self
        self.baseline_env_name = baseline_env_name
        return self

    def set_environment_name(self, environment_name):
        # type: (Self, Text) -> Self
        self.environment_name = environment_name
        return self

    def set_save_diffs(self, save_diffs):
        # type: (Self, bool) -> Self
        self.save_diffs = save_diffs
        return self

    def set_app_name(self, app_name):
        # type: (Self, Text) -> Self
        self.app_name = app_name
        return self

    def set_test_name(self, test_name):
        # type: (Self, Text) -> Self
        self.test_name = test_name
        return self

    def set_viewport_size(self, viewport_size):
        # type: (Self, ViewPort) -> Self
        self.viewport_size = viewport_size
        return self

    def set_session_type(self, session_type):
        # type: (Self, SessionType) -> Self
        self.session_type = session_type
        return self

    @property
    def ignore_caret(self):
        # type: () -> bool
        ignore = self.default_match_settings.ignore_caret
        return True if ignore is None else ignore

    def set_ignore_caret(self, ignore_caret):
        # type: (Self, bool) -> Self
        self.default_match_settings.ignore_caret = ignore_caret
        return self

    def set_host_app(self, host_app):
        # type: (Self, Text) -> Self
        self.host_app = host_app
        return self

    def set_host_os(self, host_os):
        # type: (Self, Text) -> Self
        self.host_os = host_os
        return self

    def set_match_timeout(self, match_timeout):
        # type: (Self, int) -> Self
        self.match_timeout = match_timeout
        return self

    def set_match_level(self, match_level):
        # type: (Self, MatchLevel) -> Self
        self.match_level = match_level
        return self

    def set_ignore_displacements(self, ignore_displacements):
        # type: (Self, bool) -> Self
        self.ignore_displacements = ignore_displacements
        return self

    def set_save_new_tests(self, save_new_tests):
        # type: (Self, bool) -> Self
        self.save_new_tests = save_new_tests
        return self

    def set_save_failed_tests(self, save_failed_tests):
        # type: (Self, bool) -> Self
        self.save_failed_tests = save_failed_tests
        return self

    def set_failure_reports(self, failure_reports):
        # type: (Self, FailureReports) -> Self
        self.failure_reports = failure_reports
        return self

    def set_send_dom(self, send_dom):
        # type: (Self, bool) -> Self
        self.send_dom = send_dom
        return self

    def set_use_dom(self, use_dom):
        # type: (Self, bool) -> Self
        self.use_dom = use_dom
        return self

    def set_enable_patterns(self, enable_patterns):
        # type: (Self, bool) -> Self
        self.enable_patterns = enable_patterns
        return self

    def set_stitch_overlap(self, stitch_overlap):
        # type: (Self, int) -> Self
        self.stitch_overlap = stitch_overlap
        return self

    def set_api_key(self, api_key):
        # type: (Self, Text) -> Self
        self.api_key = api_key
        return self

    def set_server_url(self, server_url):
        # type: (Self, Text) -> Self
        self.server_url = server_url
        return self

    def set_default_match_settings(self, default_match_settings):
        # type: (Self, ImageMatchSettings) -> Self
        self.default_match_settings = default_match_settings
        return self

    @property
    def accessibility_validation(self):
        # type: (Self) -> Optional[AccessibilitySettings]
        return self.default_match_settings.accessibility_settings

    @accessibility_validation.setter
    def accessibility_validation(self, accessibility_settings):
        # type: (Self, Optional[AccessibilitySettings]) -> None
        if accessibility_settings is None:
            self.self.default_match_settings.accessibility_settings = None
            return
        argument_guard.is_a(accessibility_settings, AccessibilitySettings)
        self.default_match_settings.accessibility_settings = accessibility_settings

    def set_accessibility_validation(self, accessibility_settings):
        # type: (Self, Optional[AccessibilitySettings]) -> Self
        self.accessibility_validation = accessibility_settings
        return self

    @match_timeout.validator
    def _validate1(self, attribute, value):
        if 0 < value < MINIMUM_MATCH_TIMEOUT_MS:
            raise ValueError(
                "Match timeout must be at least {} ms.".format(MINIMUM_MATCH_TIMEOUT_MS)
            )

    @viewport_size.validator
    def _validate2(self, attribute, value):
        if value is None:
            return None
        if isinstance(value, RectangleSize) or (
            isinstance(value, dict)
            and "width" in value.keys()
            and "height" in value.keys()
        ):
            return None

        raise ValueError("Wrong viewport type settled")

    @property
    def is_send_dom(self):
        # type: () -> bool
        return self.send_dom

    def clone(self):
        # type: () -> Self
        # TODO: Remove this huck when get rid of Python2
        # deepcopy on python 2 raise an exception so construct manually
        conf = copy(self)
        conf.batch = deepcopy(conf.batch)
        conf.viewport_size = copy(conf.viewport_size)
        conf.properties = deepcopy(conf.properties)
        conf.default_match_settings = deepcopy(conf.default_match_settings)
        return conf

    def add_property(self, name, value):
        # type: (Text, Text) -> Self
        """
        Associates a key/value pair with the test. This can be used later for filtering.
        :param name: (string) The property name.
        :param value: (string) The property value
        """
        self.properties.append({"name": name, "value": value})
        return self

    def clear_properties(self):
        # type: () -> Self
        """
        Clears the list of custom properties.
        """
        del self.properties[:]
        return self

    def set_proxy(self, proxy):
        # type: (Optional[ProxySettings]) -> Self
        argument_guard.is_a(proxy, ProxySettings)
        self.proxy = proxy
        return self

    @deprecated.attribute("Configuration features are not supported")
    def set_features(self, *_):
        pass



from __future__ import unicode_literals

from typing import TYPE_CHECKING, List, Optional, Text, Tuple, Union, overload

import attr
from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webelement import WebElement
from six import string_types

from applitools.common.accessibility import AccessibilityRegionType
from applitools.common.geometry import AccessibilityRegion, Region
from applitools.common.ultrafastgrid import VisualGridOption
from applitools.common.utils import argument_guard
from applitools.common.utils.json_utils import JsonInclude
from applitools.common.validators import is_list_or_tuple, is_webelement
from applitools.core.fluent import CheckSettings, CheckSettingsValues

from .region import (
    AccessibilityRegionByElement,
    AccessibilityRegionBySelector,
    FloatingRegionByElement,
    FloatingRegionBySelector,
    RegionByElement,
    RegionBySelector,
)
from .target_path import Locator, TargetPath

if TYPE_CHECKING:
    from applitools.common.utils.custom_types import (
        FLOATING_VALUES,
        AnyWebElement,
        BySelector,
        CodedRegionPadding,
        CssSelector,
        FrameIndex,
        FrameNameOrId,
    )
    from applitools.core.extract_text import OCRRegion

BEFORE_CAPTURE_SCREENSHOT = "beforeCaptureScreenshot"


@attr.s
class FrameLocator(object):
    frame_element = attr.ib(default=None)  # type: AnyWebElement
    frame_selector = attr.ib(default=None)  # type: BySelector
    frame_name_or_id = attr.ib(default=None)  # type: FrameNameOrId
    frame_index = attr.ib(default=None)  # type: FrameIndex
    scroll_root_selector = attr.ib(default=None)  # type: BySelector
    scroll_root_element = attr.ib(default=None)  # type: AnyWebElement


@attr.s
class LazyLoadOptions(object):
    scroll_length = attr.ib(default=None)
    waiting_time = attr.ib(default=None)
    page_height = attr.ib(default=None)


@attr.s
class SeleniumCheckSettingsValues(CheckSettingsValues):
    # hide_caret = attr.ib(init=False, default=None)
    scroll_root_element = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, init=False, default=None
    )  # type: WebElement
    scroll_root_selector = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, init=False, default=None
    )  # type: Locator
    target_selector = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, init=False, default=None
    )  # type: Locator
    target_element = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, init=False, default=None
    )  # type: WebElement
    frame_chain = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, init=False, factory=list
    )  # type: List[FrameLocator]

    script_hooks = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, factory=dict
    )  # type: dict
    visual_grid_options = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, default=()
    )  # type: Tuple[VisualGridOption]
    disable_browser_fetching = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, default=None
    )  # type: Optional[bool]
    ocr_region = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, init=False, default=None
    )  # type: Optional[OCRRegion]
    layout_breakpoints = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, default=None
    )  # type: Optional[Union[bool, List[int]]]
    lazy_load = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, default=None
    )  # type: Optional[LazyLoadOptions]

    @property
    def size_mode(self):
        if self.is_target_empty:
            if self.stitch_content:
                return "full-page"
            return "viewport"
        elif self.target_region:
            return "region"
        elif self.stitch_content:
            return "full-selector"
        return "selector"

    @property
    def is_target_empty(self):
        # type: () -> bool
        return (
            self.target_region is None
            and self.target_selector is None
            and self.target_element is None
        )


@attr.s
class SeleniumCheckSettings(CheckSettings):
    values = attr.ib(
        init=False, factory=SeleniumCheckSettingsValues
    )  # type: SeleniumCheckSettingsValues

    _region = attr.ib(default=None)
    _frame = attr.ib(default=None)

    @overload  # noqa
    def layout(self, *by, **kwargs):
        # type: (*BySelector,**Optional[CodedRegionPadding]) -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def layout(self, *element, **kwargs):
        # type: (*AnyWebElement,**Optional[CodedRegionPadding]) -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def layout(self, *css_selector, **kwargs):
        # type: (*CssSelector,**Optional[CodedRegionPadding]) -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def layout(self, *target_path, **kwargs):
        # type: (*Locator,**Optional[CodedRegionPadding]) -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def layout(self, *region, **kwargs):
        # type: (*Region,**Optional[CodedRegionPadding]) -> SeleniumCheckSettings
        pass

    def layout(self, *region, **kwargs):  # noqa
        """
        Adds one or more layout regions. And allow to set up paddings for them.

        :param kwargs: accepts `padding` parameter where there region paddings
         could be specified.
         Example of definition: `padding=dict(top=10, left=10, right=11, bottom=12)`
          where each keys are optional
        """
        return super(SeleniumCheckSettings, self).layout(*region, **kwargs)

    @overload  # noqa
    def strict(self, *by, **kwargs):
        # type: (*BySelector, **Optional[CodedRegionPadding]) -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def strict(self, *element, **kwargs):
        # type: (*AnyWebElement,**Optional[CodedRegionPadding]) -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def strict(self, *css_selector, **kwargs):
        # type: (*CssSelector,**Optional[CodedRegionPadding]) -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def strict(self, *target_path, **kwargs):
        # type: (*Locator,**Optional[CodedRegionPadding]) -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def strict(self, *region, **kwargs):
        # type: (*Region, **Optional[CodedRegionPadding]) -> SeleniumCheckSettings
        pass

    def strict(self, *region, **kwargs):  # noqa
        """
        Adds one or more strict regions. And allow to set up paddings for them.

        :param kwargs: accepts `padding` parameter where there region paddings
         could be specified.
         Example of definition: `padding=dict(top=10, left=10, right=11, bottom=12)`
          where each keys are optional
        """
        return super(SeleniumCheckSettings, self).strict(*region, **kwargs)

    @overload  # noqa
    def content(self, *by, **kwargs):
        # type: (*BySelector, **Optional[CodedRegionPadding]) -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def content(self, *element, **kwargs):
        # type: (*AnyWebElement,**Optional[CodedRegionPadding]) -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def content(self, *css_selector, **kwargs):
        # type: (*CssSelector, **Optional[CodedRegionPadding]) -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def content(self, *target_path, **kwargs):
        # type: (*Locator, **Optional[CodedRegionPadding]) -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def content(self, *region, **kwargs):
        # type: (*Region, **Optional[CodedRegionPadding]) -> SeleniumCheckSettings
        pass

    def content(self, *region, **kwargs):  # noqa
        """
        Adds one or more content regions. And allow to set up paddings for them.

        :param kwargs: accepts `padding` parameter where there region paddings
         could be specified.
         Example of definition: `padding=dict(top=10, left=10, right=11, bottom=12)`
          where each keys are optional
        """
        return super(SeleniumCheckSettings, self).content(*region, **kwargs)

    @overload  # noqa
    def ignore(self, *by, **kwargs):
        # type: (*BySelector, **Optional[CodedRegionPadding]) -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def ignore(self, *element, **kwargs):
        # type: (*AnyWebElement,**Optional[CodedRegionPadding]) -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def ignore(self, *css_selector, **kwargs):
        # type: (*CssSelector, **Optional[CodedRegionPadding]) -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def ignore(self, *target_path, **kwargs):
        # type: (*Locator, **Optional[CodedRegionPadding]) -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def ignore(self, *region, **kwargs):
        # type: (*Region, **Optional[CodedRegionPadding]) -> SeleniumCheckSettings
        pass

    def ignore(self, *region, **kwargs):  # noqa
        """
        Adds one or more ignore regions. And allow to set up paddings for them.

        :param kwargs: accepts `padding` parameter where there region paddings
         could be specified.
         Example of definition: `padding=dict(top=10, left=10, right=11, bottom=12)`
          where each keys are optional
        """
        return super(SeleniumCheckSettings, self).ignore(*region, **kwargs)

    @overload  # noqa
    def floating(self, max_offset, region):
        # type: (int, FLOATING_VALUES) -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def floating(
        self, region, max_up_offset, max_down_offset, max_left_offset, max_right_offset
    ):
        # type: (FLOATING_VALUES, int, int, int, int) -> SeleniumCheckSettings
        pass

    def floating(self, *args):  # noqa
        return super(SeleniumCheckSettings, self).floating(*args)

    @overload  # noqa
    def accessibility(self, region):  # noqa
        # type:(AccessibilityRegion) -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def accessibility(self, css_selector, type):  # noqa
        # type:(CssSelector, AccessibilityRegionType) -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def accessibility(self, target_path, type):  # noqa
        # type:(Locator, AccessibilityRegionType) -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def accessibility(self, by, type):  # noqa
        # type:(BySelector, AccessibilityRegionType) -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def accessibility(self, element, type):  # noqa
        # type:(AnyWebElement, AccessibilityRegionType) -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def accessibility(self, region, type):  # noqa
        # type:(Region, AccessibilityRegionType) -> SeleniumCheckSettings
        pass

    def accessibility(self, region, type=None):  # noqa
        # type:(...) -> SeleniumCheckSettings
        return super(SeleniumCheckSettings, self).accessibility(region, type)

    @overload  # noqa
    def region(self, region):
        # type: (Region) -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def region(self, css_selector):
        # type: (CssSelector) -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def region(self, target_path):
        # type: (Locator) -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def region(self, element):
        # type: (AnyWebElement) -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def region(self, by):
        # type: (BySelector) -> SeleniumCheckSettings
        pass

    def region(self, region):  # noqa
        # type:(...) -> SeleniumCheckSettings
        if isinstance(region, Region):
            self.values.target_region = region
        elif is_list_or_tuple(region):
            self.values.target_selector = TargetPath.region(*region)
        elif isinstance(region, string_types):
            self.values.target_selector = TargetPath.region(region)
        elif isinstance(region, Locator):
            self.values.target_selector = region
        elif is_webelement(region):
            self.values.target_element = region
        else:
            raise TypeError("region method called with argument of unknown type!")
        return self

    @overload  # noqa
    def frame(self, frame_name_or_id):
        # type: (FrameNameOrId) -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def frame(self, element):
        # type: (AnyWebElement) -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def frame(self, index):
        # type: (FrameIndex) -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def frame(self, by):
        # type: (BySelector) -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def frame(self, target_path):
        # type: (Locator) -> SeleniumCheckSettings
        pass

    def frame(self, frame):  # noqa
        # type:(...) -> SeleniumCheckSettings
        fl = FrameLocator()
        if isinstance(frame, int):
            fl.frame_index = frame
        elif isinstance(frame, string_types):
            fl.frame_name_or_id = frame
        elif is_webelement(frame):
            fl.frame_element = frame
        elif is_list_or_tuple(frame):
            fl.frame_selector = TargetPath.frame(*frame)
        elif isinstance(frame, Locator):
            fl.frame_selector = frame
        else:
            raise TypeError("frame method called with argument of unknown type!")
        self.values.frame_chain.append(fl)
        return self

    def before_render_screenshot_hook(self, hook):
        # type: (Text) -> SeleniumCheckSettings
        self.values.script_hooks[BEFORE_CAPTURE_SCREENSHOT] = hook
        return self

    def _region_provider_from(self, region, method_name, padding):
        if isinstance(region, string_types):
            return RegionBySelector(TargetPath.region(region), padding)
        elif is_list_or_tuple(region):
            return RegionBySelector(TargetPath.region(*region), padding)
        elif isinstance(region, Locator):
            return RegionBySelector(region, padding)
        elif is_webelement(region):
            return RegionByElement(region, padding)
        return super(SeleniumCheckSettings, self)._region_provider_from(
            region, method_name, padding
        )

    def _set_scroll_root_selector(self, target_path):
        if len(self.values.frame_chain) == 0:
            self.values.scroll_root_selector = target_path
        else:
            self.values.frame_chain[-1].scroll_root_selector = target_path

    def _set_scroll_root_element(self, element):
        if len(self.values.frame_chain) == 0:
            self.values.scroll_root_element = element
        else:
            self.values.frame_chain[-1].scroll_root_element = element

    @overload  # noqa
    def scroll_root_element(self, element):
        # type: (AnyWebElement) -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def scroll_root_element(self, selector):
        # type: (CssSelector) -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def scroll_root_element(self, target_path):
        # type: (Locator) -> SeleniumCheckSettings
        pass

    @overload  # noqa
    def scroll_root_element(self, by):
        # type: (BySelector) -> SeleniumCheckSettings
        pass

    def scroll_root_element(self, element_or_selector):  # noqa
        if isinstance(element_or_selector, string_types):
            self._set_scroll_root_selector(TargetPath.region(element_or_selector))
        elif is_list_or_tuple(element_or_selector):
            self._set_scroll_root_selector(TargetPath.region(*element_or_selector))
        elif isinstance(element_or_selector, Locator):
            self._set_scroll_root_selector(element_or_selector)
        elif is_webelement(element_or_selector):
            self._set_scroll_root_element(element_or_selector)
        else:
            raise TypeError("Unsupported type of scroll root element")
        return self

    def _floating_provider_from(self, region, bounds):
        if is_webelement(region):
            return FloatingRegionByElement(region, bounds)
        elif isinstance(region, string_types):
            return FloatingRegionBySelector(TargetPath.region(region), bounds)
        elif is_list_or_tuple(region):
            return FloatingRegionBySelector(TargetPath.region(*region), bounds)
        elif isinstance(region, Locator):
            return FloatingRegionBySelector(region, bounds)
        return super(SeleniumCheckSettings, self)._floating_provider_from(
            region, bounds
        )

    def _accessibility_provider_from(self, region, accessibility_region_type):
        if is_webelement(region):
            return AccessibilityRegionByElement(region, accessibility_region_type)
        elif isinstance(region, string_types):
            return AccessibilityRegionBySelector(
                TargetPath.region(region), accessibility_region_type
            )
        elif is_list_or_tuple(region):
            return AccessibilityRegionBySelector(
                TargetPath.region(*region), accessibility_region_type
            )
        elif isinstance(region, Locator):
            return AccessibilityRegionBySelector(region, accessibility_region_type)
        return super(SeleniumCheckSettings, self)._accessibility_provider_from(
            region, accessibility_region_type
        )

    def visual_grid_options(self, *options):
        # type: (*VisualGridOption) -> SeleniumCheckSettings
        argument_guard.are_(options, VisualGridOption)
        self.values.visual_grid_options = options
        return self

    def disable_browser_fetching(self, disable=True):
        # type: (bool) -> SeleniumCheckSettings
        self.values.disable_browser_fetching = disable
        return self

    @overload
    def layout_breakpoints(self, enabled):
        # type: (bool) -> SeleniumCheckSettings
        pass

    @overload
    def layout_breakpoints(self, *breakpoints):
        # type: (*int) -> SeleniumCheckSettings
        pass

    def layout_breakpoints(self, enabled_or_first, *rest):
        if isinstance(enabled_or_first, bool):
            self.values.layout_breakpoints = enabled_or_first
        elif isinstance(enabled_or_first, int):
            self.values.layout_breakpoints = [enabled_or_first] + list(rest)
        else:
            raise TypeError(
                "{} is not an instance of bool or int".format(enabled_or_first)
            )
        return self

    def lazy_load(self, scroll_length=300, waiting_time=2000, page_height=15000):
        # type: (int, int, int) -> SeleniumCheckSettings
        lazy_load = LazyLoadOptions(scroll_length, waiting_time, page_height)
        self.values.lazy_load = lazy_load
        return self

    @property
    def is_check_window(self):
        # type: () -> bool
        return self.values.is_target_empty



from typing import TYPE_CHECKING, List, Optional, Text, TypeVar, overload

import attr
from six import raise_from

from applitools.common import FloatingBounds, MatchLevel
from applitools.common.accessibility import AccessibilityRegionType
from applitools.common.geometry import AccessibilityRegion, Rectangle, Region
from applitools.common.utils import argument_guard
from applitools.common.utils.json_utils import JsonInclude

from .region import (
    AccessibilityRegionByRectangle,
    FloatingRegionByRectangle,
    GetAccessibilityRegion,
    GetFloatingRegion,
    GetRegion,
    RegionByRectangle,
)

if TYPE_CHECKING:
    from applitools.common.utils.custom_types import CodedRegionPadding, Num

__all__ = ("CheckSettings", "CheckSettingsValues")


@attr.s
class CheckSettingsValues(object):
    """
    Access to values stored in :py:class:`CheckSettings`
    """

    target_region = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, init=False, default=None
    )  # type: Optional[Region]
    timeout = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, init=False, default=-1
    )  # type: Num  # milliseconds

    ignore_caret = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, init=False, default=None
    )  # type: Optional[bool]
    stitch_content = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, init=False, default=None
    )  # type: Optional[bool]
    match_level = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, init=False, default=None
    )  # type: Optional[MatchLevel]
    name = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, init=False, default=None
    )  # type: Optional[Text]

    send_dom = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, init=False, default=None
    )  # type: Optional[bool]
    use_dom = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, init=False, default=None
    )  # type: Optional[bool]
    enable_patterns = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, init=False, default=None
    )  # type: Optional[bool]
    ignore_displacements = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, init=False, default=None
    )  # type: Optional[bool]

    ignore_regions = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, init=False, factory=list
    )  # type: List[GetRegion]
    layout_regions = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, init=False, factory=list
    )  # type: List[GetRegion]
    strict_regions = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, init=False, factory=list
    )  # type: List[GetRegion]
    content_regions = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, init=False, factory=list
    )  # type: List[GetRegion]
    floating_regions = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, init=False, factory=list
    )  # type: List[GetFloatingRegion]
    accessibility_regions = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, init=False, factory=list
    )  # type: List[GetAccessibilityRegion]
    variation_group_id = attr.ib(
        metadata={JsonInclude.NON_NONE: True}, init=False, default=None
    )  # type: Optional[Text]


Self = TypeVar("Self", bound="CheckSettings")  # typedef


@attr.s
class CheckSettings(object):
    """
    The Match settings object to use in the various Eyes.Check methods.
    """

    values = attr.ib(
        init=False, factory=CheckSettingsValues
    )  # type: CheckSettingsValues

    def variation_group_id(self, variation_group_id):
        # type: (Text) -> Self
        self.values.variation_group_id = variation_group_id
        return self

    def layout(self, *regions, **kwargs):
        # type: (Self, *Region, **Optional[CodedRegionPadding])  -> Self
        """Shortcut to set the match level to :py:attr:`MatchLevel.LAYOUT`."""
        if not regions:
            self.values.match_level = MatchLevel.LAYOUT
            return self
        try:
            self.values.layout_regions = self.__regions(
                regions, method_name="layout_regions", padding=kwargs.get("padding")
            )
        except TypeError as e:
            raise_from(TypeError("Wrong argument in .layout()"), e)
        return self

    def exact(self):
        """Shortcut to set the match level to :py:attr:`MatchLevel.EXACT` if no args"""
        self.values.match_level = MatchLevel.EXACT
        return self

    def strict(self, *regions, **kwargs):
        # type: (Self, *Region, **Optional[CodedRegionPadding])  -> Self
        """Shortcut to set the match level to :py:attr:`MatchLevel.STRICT` if no args"""
        if not regions:
            self.values.match_level = MatchLevel.STRICT
            return self
        try:
            self.values.strict_regions = self.__regions(
                regions, method_name="strict_regions", padding=kwargs.get("padding")
            )
        except TypeError as e:
            raise_from(TypeError("Wrong argument in .strict()"), e)
        return self

    def content(self, *regions, **kwargs):
        # type: (Self, *Region, **Optional[CodedRegionPadding])  -> Self
        """Shortcut to set the match level to :py:attr:`MatchLevel.CONTENT` if no args"""
        if not regions:
            self.values.match_level = MatchLevel.CONTENT
            return self
        try:
            self.values.content_regions = self.__regions(
                regions, method_name="content_regions", padding=kwargs.get("padding")
            )
        except TypeError as e:
            raise_from(TypeError("Wrong argument in .content()"), e)
        return self

    def ignore(self, *regions, **kwargs):
        # type: (Self, *Region, **Optional[CodedRegionPadding])  -> Self
        """Adds one or more ignore regions."""
        try:
            self.values.ignore_regions = self.__regions(
                regions, method_name="ignore_regions", padding=kwargs.get("padding")
            )
        except TypeError as e:
            raise_from(TypeError, e)
        return self

    @overload  # noqa
    def accessibility(self, region):
        # type:(Self, AccessibilityRegion) -> Self
        pass

    @overload  # noqa
    def accessibility(self, region, type):
        # type:(Self, Region, AccessibilityRegionType) -> Self
        pass

    def accessibility(self, region, type=None):  # noqa
        """Adds one accessibility region."""
        if type:
            argument_guard.is_a(type, AccessibilityRegionType)
        try:
            self.values.accessibility_regions.append(
                self._accessibility_provider_from(region, type)
            )
        except TypeError as e:
            raise_from(TypeError("Wrong argument in .accessibility()"), e)
        return self

    @overload  # noqa
    def floating(self, max_offset, region):
        # type: (Self, int, Region) -> Self
        pass

    @overload  # noqa
    def floating(
        self, region, max_up_offset, max_down_offset, max_left_offset, max_right_offset
    ):
        # type: (Self, Region, int, int, int, int) -> Self
        pass

    def floating(self, *args):  # noqa
        """
        Adds a floating region. Region and max_offset or [max_up_offset, max_down_offset, "
                "max_left_offset, max_right_offset] are required parameters.

        :param arg1: max_offset | Region
        :param arg2: Region     | max_up_offset
        :param arg3: None       | max_down_offset
        :param arg4: None       | max_left_offset
        :param arg5: None       | max_right_offset
        """
        if len(args) < 2:
            raise TypeError("Not enough arguments")
        if isinstance(args[0], int) and not isinstance(args[1], int):
            max_offset = args[0]  # type: int
            region = args[1]  # type: ignore
            bounds = FloatingBounds(
                max_up_offset=max_offset,
                max_down_offset=max_offset,
                max_left_offset=max_offset,
                max_right_offset=max_offset,
            )
        elif (
            isinstance(args[1], int)
            and isinstance(args[2], int)
            and isinstance(args[3], int)
            and isinstance(args[4], int)
        ):
            region = args[0]  # type: ignore
            bounds = FloatingBounds(
                max_up_offset=args[1],
                max_down_offset=args[2],
                max_left_offset=args[3],
                max_right_offset=args[4],
            )
        else:
            raise TypeError("No type match")
        try:
            region_or_container = self._floating_provider_from(region, bounds)
        except TypeError as e:
            raise_from(TypeError("Wrong arguments in .floating()"), e)
        self.values.floating_regions.append(region_or_container)
        return self

    def send_dom(self, senddom=True):
        # type: (Self, bool) -> Self
        """
        Defines whether to send the document DOM or not.
        """
        self.values.send_dom = senddom
        return self

    def use_dom(self, use=True):
        # type: (Self, bool) -> Self
        """
        Defines useDom for enabling the match algorithm to use dom.
        """
        self.values.use_dom = use
        return self

    def enable_patterns(self, enable=True):
        # type: (Self, bool) -> Self
        self.values.enable_patterns = enable
        return self

    def ignore_displacements(self, should_ignore=True):
        # type: (Self, bool) -> Self
        self.values.ignore_displacements = should_ignore
        return self

    def match_level(self, match_level):
        # type: (Self, MatchLevel)  -> Self
        self.values.match_level = match_level
        return self

    def ignore_caret(self, ignore=True):
        # type: (Self, bool)  -> Self
        self.values.ignore_caret = ignore
        return self

    def fully(self, fully=True):
        # type: (Self, bool)  -> Self
        self.values.stitch_content = fully
        return self

    def with_name(self, name):
        # type: (Self, Text)  -> Self
        self.values.name = name
        return self

    def stitch_content(self, stitch_content=True):
        # type: (Self, bool)  -> Self
        self.values.stitch_content = stitch_content
        return self

    def timeout(self, timeout):
        # type: (Self, int)  -> Self
        self.values.timeout = timeout
        return self

    def __regions(self, regions, method_name, padding):
        if not regions:
            raise TypeError(
                "{name} method called without arguments!".format(name=method_name)
            )

        regions_list = getattr(self.values, method_name)
        for region in regions:
            regions_list.append(
                self._region_provider_from(region, method_name, padding)
            )
        return regions_list

    def _region_provider_from(self, region, method_name, padding):
        if isinstance(region, Region):
            return RegionByRectangle(region)
        raise TypeError(
            "Unsupported region: \n\ttype: {} \n\tvalue: {}".format(
                type(region), region
            )
        )

    def _floating_provider_from(self, region, bounds):
        if isinstance(region, Region):
            return FloatingRegionByRectangle(Region.from_(region), bounds)
        raise TypeError(
            "Unsupported floating region: \n\ttype: {} \n\tvalue: {}".format(
                type(region), region
            )
        )

    def _accessibility_provider_from(self, region, accessibility_region_type):
        if isinstance(region, (Region, Rectangle)) and accessibility_region_type:
            return AccessibilityRegionByRectangle(region, accessibility_region_type)
        elif isinstance(region, AccessibilityRegion):
            return AccessibilityRegionByRectangle(region)
        raise TypeError(
            "Unsupported accessibility region: \n\ttype: {} \n\tvalue: {}".format(
                type(region), region
            )
        )



import typing

import attr

from applitools.common.accessibility import AccessibilityRegionType
from applitools.core.fluent.region import (
    GetAccessibilityRegion,
    GetFloatingRegion,
    GetRegion,
)

if typing.TYPE_CHECKING:
    from typing import Optional

    from applitools.common import FloatingBounds
    from applitools.common.utils.custom_types import AnyWebElement, CodedRegionPadding
    from applitools.selenium.fluent.target_path import Locator

__all__ = (
    "RegionBySelector",
    "RegionByElement",
    "FloatingRegionBySelector",
    "FloatingRegionByElement",
    "AccessibilityRegionBySelector",
    "AccessibilityRegionByElement",
)


class GetSeleniumRegion(GetRegion):
    pass


@attr.s
class RegionByElement(GetSeleniumRegion):
    _element = attr.ib()  # type: AnyWebElement
    _padding = attr.ib(default=None)  # type: Optional[CodedRegionPadding]


@attr.s
class RegionBySelector(GetSeleniumRegion):
    """
    :param by: The "by" part of a selenium selector for an element which
               represents the ignore region
    :param value: The "value" part of a selenium selector for
                  an element which represents the ignore region.
    """

    _target_path = attr.ib()  # type: Locator
    _padding = attr.ib(default=None)  # type: Optional[CodedRegionPadding]


class GetSeleniumFloatingRegion(GetFloatingRegion):
    pass


@attr.s
class FloatingRegionByElement(GetSeleniumFloatingRegion):
    """
    :ivar element: The element which represents the inner region (the floating part).
    :ivar bounds: The outer rectangle bounding the inner region.
    """

    _element = attr.ib()  # type: AnyWebElement
    _bounds = attr.ib()  # type: FloatingBounds


@attr.s
class FloatingRegionBySelector(GetSeleniumFloatingRegion):
    """
    :ivar by: The selenium By
    :ivar value: The css selector for an element which represents the inner region.
    :ivar bounds: The outer rectangle bounding the inner region.
    """

    _target_path = attr.ib()  # type: Locator
    _bounds = attr.ib()  # type: FloatingBounds


class GetSeleniumAccessibilityRegion(GetAccessibilityRegion):
    pass


@attr.s
class AccessibilityRegionBySelector(GetSeleniumAccessibilityRegion):
    _target_path = attr.ib()  # type: Locator
    _type = attr.ib()  # type: AccessibilityRegionType


@attr.s
class AccessibilityRegionByElement(GetSeleniumAccessibilityRegion):
    _element = attr.ib()  # type: AnyWebElement
    _type = attr.ib()  # type: AccessibilityRegionType



"""
Contains the Trigger classes (e.g., click, text writing) used by Eyes.
"""
from __future__ import absolute_import

import typing as tp
from abc import ABCMeta
from collections import OrderedDict

from six import add_metaclass

from applitools.common.errors import EyesError

if tp.TYPE_CHECKING:
    from applitools.common.geometry import Point, Region

__all__ = ("TextTrigger", "MouseTrigger")

MOUSE_ACTION = {
    "click": "Click",
    "right_click": "RightClick",
    "double_click": "DoubleClick",
    "move": "Move",
    "down": "Down",
    "up": "Up",
}

_TRIGGER_TYPE_FIELD_NAME = "triggerType"
_TRIGGER_TYPE = {
    "unknown": "Unknown",
    "mouse": "Mouse",
    "text": "Text",
    "keyboard": "Keyboard",
}


@add_metaclass(ABCMeta)
class ActionTrigger(object):
    pass


class MouseTrigger(ActionTrigger):
    """Encapsulates a mouse trigger."""

    def __init__(self, action, control, location):
        # type: (tp.Text, Region, Point) -> None
        """
        Ctor.

        :param action: The mouse action represented by the trigger.
        :param control: The region to which the text was entered.
        :param location: The location which was clicked relative to the control.
        """
        self.action = action
        self.control = control
        self.location = location  # Point. Relative to the control's (left,top) corner.

    def __getstate__(self):
        return OrderedDict(
            [
                (_TRIGGER_TYPE_FIELD_NAME, _TRIGGER_TYPE["mouse"]),
                ("mouseAction", MOUSE_ACTION[self.action]),
                ("control", self.control),
                ("location", self.location),
            ]
        )

    # Required is required in order for jsonpickle to work on this object.
    # noinspection PyMethodMayBeStatic
    def __setstate__(self, state):
        raise EyesError("Cannot create MouseTrigger instance from dict!")

    def __str__(self):
        return "%s [%s] %s" % (self.action, self.control, self.location)


class TextTrigger(ActionTrigger):
    """Encapsulates a text input by the user."""

    def __init__(self, control, text):
        # type: (Region, tp.Text) -> None
        """
        Ctor.

        :param control: The region to which the text was entered.
        :param text: The trigger's text.
        """
        self.control = control
        self.text = text

    def __getstate__(self):
        return OrderedDict(
            [
                (_TRIGGER_TYPE_FIELD_NAME, _TRIGGER_TYPE["text"]),
                ("control", self.control),
                ("text", self.text),
            ]
        )

    # Required is required in order for jsonpickle to work on this object.
    # noinspection PyMethodMayBeStatic
    def __setstate__(self, state):
        raise EyesError("Cannot create MouseTrigger instance from dict!")

    def __str__(self):
        return "Text [%s] %s" % (self.control, self.text)



from __future__ import absolute_import

import os
import typing

"""
General purpose utilities.
"""


if typing.TYPE_CHECKING:
    from typing import Any, Callable, List, Optional, Text

    T = typing.TypeVar("T")


def cached_property(f):
    # type: (Callable) -> Any
    """
    Returns a cached property that is calculated by function f
    """

    def get(self):
        try:
            return self._property_cache[f]
        except AttributeError:
            self._property_cache = {}
            x = self._property_cache[f] = f(self)
            return x
        except KeyError:
            x = self._property_cache[f] = f(self)
            return x

    return property(get)


def get_env_with_prefix(env_name, default=None):
    # type: (Text, Optional[Text]) -> Optional[Text]
    """
    Takes name of ENV variable, check if exists origin and with list of prefixes
    """
    prefixes_to_check = ["bamboo"]
    try:
        return os.environ[env_name]
    except KeyError:
        for prefix in prefixes_to_check:
            name = "{}_{}".format(prefix, env_name)
            value = os.getenv(name)
            if value:
                return value
    return default


class DynamicEnumGetter(object):
    """
    Allow to use methods as values in Enum classes
    """

    def __init__(self, fget=None):
        self.fget = fget
        self.__doc__ = fget.__doc__

    def __get__(self, instance, ownerclass=None):
        return self.fget(ownerclass)



import typing as tp
import warnings
from functools import wraps
from inspect import getcallargs


def module(name, recommendation):
    # type: (tp.Text, tp.Text) -> None
    message = "Module {} is deprecated: {}".format(name, recommendation)
    warnings.warn(message, stacklevel=2, category=ImportWarning)


def argument(name, recommendation):
    # type: (tp.Text, tp.Text) -> tp.Callable[[tp.Callable], tp.Callable]
    def wrapper(func):
        @wraps(func)
        def wrapped(*args, **kwargs):
            call_args = getcallargs(func, *args, **kwargs)
            argval = call_args.get(name)
            if argval is not None:
                message = "Argument {} of {} is deprecated: {}".format(
                    name, func.__name__, recommendation
                )
                warnings.warn(message, stacklevel=2, category=DeprecationWarning)
            return func(*args, **kwargs)

        return wrapped

    return wrapper


def attribute(recommendation):
    # type: (tp.Text) -> tp.Callable[[tp.Callable], tp.Callable]
    def wrapper(attr):
        @wraps(attr)
        def wrapped(*args, **kwargs):
            msg = "Use of {} is deprecated: {}".format(attr.__name__, recommendation)
            warnings.warn(msg, stacklevel=2, category=DeprecationWarning)
            return attr(*args, **kwargs)

        return wrapped

    return wrapper


warnings.filterwarnings("default", module="applitools")



"""
Stub module that that supported original log handling.
"""
from applitools.common import deprecated

__all__ = ("StdoutLogger", "FileLogger")


def StdoutLogger(*args, **kwargs):
    """Stub function to avoid breaking existing code"""
    pass


def FileLogger(*args, **kwargs):
    """Stub function to avoid breaking existing code"""
    pass


@deprecated.attribute("logging configuration is not supported anymore")
def set_logger(logger=None):
    """Stub function to avoid breaking existing code"""
    pass



from enum import Enum

import attr

from applitools.common import deprecated
from applitools.common.utils.json_utils import JsonInclude


class AccessibilityGuidelinesVersion(Enum):
    WCAG_2_0 = "WCAG_2_0"
    WCAG_2_1 = "WCAG_2_1"


class AccessibilityLevel(Enum):
    AA = "AA"
    AAA = "AAA"


@attr.s(init=False)
class AccessibilitySettings(object):
    level = attr.ib(
        type=AccessibilityLevel, metadata={JsonInclude.THIS: True}
    )  # type: AccessibilityLevel
    guidelines_version = attr.ib(
        type=AccessibilityGuidelinesVersion, metadata={JsonInclude.NAME: "version"}
    )  # type: AccessibilityGuidelinesVersion

    def __init__(
        self,
        level,  # type: AccessibilityLevel
        guidelines_version,  # type: AccessibilityGuidelinesVersion
    ):
        # type: (...) -> None
        self.level = AccessibilityLevel(level)
        self.guidelines_version = AccessibilityGuidelinesVersion(guidelines_version)


class AccessibilityRegionType(Enum):
    IgnoreContrast = "IgnoreContrast"
    RegularText = "RegularText"
    LargeText = "LargeText"
    BoldText = "BoldText"
    GraphicalObject = "GraphicalObject"


class AccessibilityStatus(Enum):
    """
    Accessibility status.
    """

    # Session has passed accessibility validation.
    Passed = "Passed"
    # Session hasn't passed accessibility validation.
    Failed = "Failed"


@attr.s(init=False)
class SessionAccessibilityStatus(object):
    level = attr.ib(
        type=AccessibilityLevel, metadata={JsonInclude.THIS: True}
    )  # type: AccessibilityLevel
    version = attr.ib(
        type=AccessibilityGuidelinesVersion, metadata={JsonInclude.THIS: True}
    )  # type: AccessibilityGuidelinesVersion
    status = attr.ib(
        type=AccessibilityStatus, metadata={JsonInclude.THIS: True}
    )  # type: AccessibilityStatus

    def __init__(
        self,
        status,  # type: AccessibilityStatus
        level,  # type: AccessibilityLevel
        version,  # type: AccessibilityGuidelinesVersion
    ):
        self.level = AccessibilityLevel(level)
        self.version = AccessibilityGuidelinesVersion(version)
        self.status = AccessibilityStatus(status)

    @property
    @deprecated.attribute("use `version` instead")
    def guidelines_version(self):
        # type: () -> AccessibilityGuidelinesVersion
        return self.version



from __future__ import absolute_import

from enum import Enum
from typing import TYPE_CHECKING, List, Optional, Text, Tuple

import attr

from applitools.common.utils.json_utils import JsonInclude

from .accessibility import SessionAccessibilityStatus
from .geometry import RectangleSize
from .match import ImageMatchSettings

if TYPE_CHECKING:
    from . import ProxySettings
    from .ultrafastgrid import RenderBrowserInfo


__all__ = ("TestResults", "TestResultsSummary", "TestResultContainer")


class TestResultsStatus(Enum):
    """
    Status values for tests results.
    """

    Passed = "Passed"
    Unresolved = "Unresolved"
    Failed = "Failed"

    __test__ = False  # avoid warnings in test frameworks


@attr.s
class SessionUrls(object):
    batch = attr.ib(default=None, metadata={JsonInclude.THIS: True})  # type: Text
    session = attr.ib(default=None, metadata={JsonInclude.THIS: True})  # type: Text


@attr.s
class StepInfo(object):
    @attr.s(repr_ns="StepInfo")
    class AppUrls(object):
        step = attr.ib(default=None, metadata={JsonInclude.THIS: True})  # type: Text
        step_editor = attr.ib(
            default=None, metadata={JsonInclude.THIS: True}
        )  # type: Text

    @attr.s(repr_ns="StepInfo")
    class ApiUrls(object):
        baseline_image = attr.ib(
            default=None, metadata={JsonInclude.THIS: True}
        )  # type: Text
        current_image = attr.ib(
            default=None, metadata={JsonInclude.THIS: True}
        )  # type: Text
        diff_image = attr.ib(
            default=None, metadata={JsonInclude.THIS: True}
        )  # type: Text
        checkpoint_image = attr.ib(
            default=None, metadata={JsonInclude.THIS: True}
        )  # type: Text
        checkpoint_image_thumbnail = attr.ib(
            default=None, metadata={JsonInclude.THIS: True}
        )  # type: Text

    name = attr.ib(default=None, metadata={JsonInclude.THIS: True})  # type: Text
    is_different = attr.ib(
        default=None, metadata={JsonInclude.THIS: True}
    )  # type: bool
    has_baseline_image = attr.ib(
        default=None, metadata={JsonInclude.THIS: True}
    )  # type: bool
    has_current_image = attr.ib(
        default=None, metadata={JsonInclude.THIS: True}
    )  # type: bool
    has_checkpoint_image = attr.ib(
        default=None, metadata={JsonInclude.THIS: True}
    )  # type: bool
    api_urls = attr.ib(
        default=None, type=ApiUrls, metadata={JsonInclude.THIS: True}
    )  # type: ApiUrls
    app_urls = attr.ib(
        default=None, type=AppUrls, metadata={JsonInclude.THIS: True}
    )  # type: AppUrls


@attr.s
class TestResults(object):
    """
    Eyes test results.
    """

    steps = attr.ib(default=0, metadata={JsonInclude.THIS: True})  # type: int
    matches = attr.ib(default=0, metadata={JsonInclude.THIS: True})  # type: int
    mismatches = attr.ib(default=0, metadata={JsonInclude.THIS: True})  # type: int
    missing = attr.ib(default=0, metadata={JsonInclude.THIS: True})  # type: int
    exact_matches = attr.ib(
        default=0, repr=False, metadata={JsonInclude.THIS: True}
    )  # type: int
    strict_matches = attr.ib(
        default=0, repr=False, metadata={JsonInclude.THIS: True}
    )  # type: int
    content_matches = attr.ib(
        default=0, repr=False, metadata={JsonInclude.THIS: True}
    )  # type: int
    layout_matches = attr.ib(
        default=0, repr=False, metadata={JsonInclude.THIS: True}
    )  # type: int
    none_matches = attr.ib(
        default=0, repr=False, metadata={JsonInclude.THIS: True}
    )  # type: int
    is_new = attr.ib(
        default=None, repr=False, metadata={JsonInclude.THIS: True}
    )  # type: Optional[bool]

    url = attr.ib(
        default=None, metadata={JsonInclude.THIS: True}
    )  # type: Optional[Text]
    status = attr.ib(
        converter=lambda v: TestResultsStatus(v) if v else None,
        type=TestResultsStatus,
        default=None,
        repr=False,
        metadata={JsonInclude.THIS: True},
    )  # type: Optional[TestResultsStatus]

    name = attr.ib(
        default=None, repr=False, metadata={JsonInclude.THIS: True}
    )  # type: Text
    secret_token = attr.ib(
        default=None, repr=False, metadata={JsonInclude.THIS: True}
    )  # type: Text
    id = attr.ib(
        default=None, repr=False, metadata={JsonInclude.THIS: True}
    )  # type: Text
    app_name = attr.ib(
        default=None, repr=False, metadata={JsonInclude.THIS: True}
    )  # type: Text
    batch_name = attr.ib(
        default=None, repr=False, metadata={JsonInclude.THIS: True}
    )  # type: Text
    batch_id = attr.ib(
        default=None, repr=False, metadata={JsonInclude.THIS: True}
    )  # type: Text
    branch_name = attr.ib(
        default=None, repr=False, metadata={JsonInclude.THIS: True}
    )  # type: Text
    host_os = attr.ib(
        default=None, repr=False, metadata={JsonInclude.THIS: True}
    )  # type: Text
    host_app = attr.ib(
        default=None, repr=False, metadata={JsonInclude.THIS: True}
    )  # type: Text
    host_display_size = attr.ib(
        default=None, repr=False, type=RectangleSize, metadata={JsonInclude.THIS: True}
    )  # type: RectangleSize
    started_at = attr.ib(
        default=None, repr=False, metadata={JsonInclude.THIS: True}
    )  # type: Text
    duration = attr.ib(
        default=None, repr=False, metadata={JsonInclude.THIS: True}
    )  # type: int
    is_different = attr.ib(
        default=None, repr=False, metadata={JsonInclude.THIS: True}
    )  # type: bool
    is_aborted = attr.ib(
        default=None, repr=False, metadata={JsonInclude.THIS: True}
    )  # type: bool
    is_empty = attr.ib(
        default=None, repr=False, metadata={JsonInclude.THIS: True}
    )  # type: bool
    app_urls = attr.ib(
        default=None, repr=False, type=SessionUrls, metadata={JsonInclude.THIS: True}
    )  # type: SessionUrls
    api_urls = attr.ib(
        default=None, repr=False, type=SessionUrls, metadata={JsonInclude.THIS: True}
    )  # type: SessionUrls
    steps_info = attr.ib(
        default=None, repr=False, type=StepInfo, metadata={JsonInclude.THIS: True}
    )  # type: List[StepInfo]
    baseline_id = attr.ib(
        default=None, repr=False, metadata={JsonInclude.THIS: True}
    )  # type: Text
    default_match_settings = attr.ib(
        default=None,
        repr=False,
        type=ImageMatchSettings,
        metadata={JsonInclude.THIS: True},
    )  # type: ImageMatchSettings
    accessibility_status = attr.ib(
        default=None,
        type=SessionAccessibilityStatus,
        metadata={JsonInclude.THIS: True},
    )  # type: SessionAccessibilityStatus
    _connection_config = attr.ib(
        default=(None, None, None),
        eq=False,
        order=False,
        metadata={JsonInclude.THIS: False},
        repr=False,
    )  # type: Tuple[Optional[Text], Optional[Text], Optional[ProxySettings]]
    __test__ = False  # avoid warnings in test frameworks

    @property
    def is_passed(self):
        # type: () -> bool
        return self.status == TestResultsStatus.Passed

    @property
    def is_unresolved(self):
        # type: () -> bool
        return self.status == TestResultsStatus.Unresolved

    @property
    def is_failed(self):
        # type: () -> bool
        return self.status == TestResultsStatus.Failed

    def set_connection_config(self, server_url, api_key, proxy_settings):
        # type: (Text, Text, Optional[ProxySettings]) -> None
        self._connection_config = (server_url, api_key, proxy_settings)

    def delete(self):
        # type: () -> None
        from applitools.selenium.__version__ import __version__
        from applitools.selenium.command_executor import CommandExecutor
        from applitools.selenium.eyes import EyesRunner
        from applitools.selenium.universal_sdk_types import marshal_delete_test_settings

        cmd = CommandExecutor.get_instance(EyesRunner.BASE_AGENT_ID, __version__)
        marshaled = marshal_delete_test_settings(self)
        cmd.core_delete_test(marshaled)


@attr.s(repr=False, str=False)
class TestResultContainer(object):
    test_results = attr.ib()  # type: TestResults
    browser_info = attr.ib()  # type: Optional[RenderBrowserInfo]
    exception = attr.ib()  # type: Optional[Exception]

    def __str__(self):
        browser_info = (
            "\n browser_info = {}".format(self.browser_info)
            if self.browser_info
            else ""
        )
        return (
            "TestResultContainer{{"
            "\n test_results = {test_results}"
            "{browser_info}"
            "\n exception = {exception}"
            "}}".format(
                test_results=self.test_results,
                browser_info=browser_info,
                exception=self.exception,
            )
        )


@attr.s(repr=False, str=False)
class TestResultsSummary(object):
    _all_results = attr.ib()  # type: List[TestResultContainer]
    _exceptions = attr.ib(init=False, default=0)  # type: int
    _passed = attr.ib(init=False, default=0)  # type: int
    _unresolved = attr.ib(init=False, default=0)  # type: int
    _failed = attr.ib(init=False, default=0)  # type: int
    _mismatches = attr.ib(init=False, default=0)  # type: int
    _missing = attr.ib(init=False, default=0)  # type: int
    _matches = attr.ib(init=False, default=0)  # type: int

    @property
    def all_results(self):
        # type: () -> List[TestResultContainer]
        return self._all_results

    def size(self):
        # type: () -> int
        return len(self)

    def __attrs_post_init__(self):
        for result_container in self._all_results:
            if result_container and result_container.exception:
                self._exceptions += 1

            result = result_container.test_results
            if result is None:
                continue
            if result.is_failed:
                self._failed += 1
            elif result.is_passed:
                self._passed += 1
            elif result.is_unresolved:
                self._unresolved += 1
            self._matches += result.matches
            self._missing += result.missing
            self._mismatches += result.mismatches

    def __len__(self):
        return len(self._all_results)

    def __iter__(self):
        return iter(self._all_results)

    def __getitem__(self, item):
        # type: (int) -> TestResultContainer
        return self._all_results[item]

    def __str__(self):
        return """
result summary {{
    all results={all_results}
    passed={passed}
    unresolved={unresolved}
    failed={failed}
    exceptions={exceptions}
    mismatches={mismatches}
    missing={missing}
    matches={matches}
}}""".format(
            all_results="\n".join(map(str, self._all_results)),
            passed=self._passed,
            unresolved=self._unresolved,
            failed=self._failed,
            exceptions=self._exceptions,
            mismatches=self._mismatches,
            missing=self._missing,
            matches=self._matches,
        )



import json
import re
from collections import Counter, defaultdict
from typing import TYPE_CHECKING

import attr
from six import iteritems

if TYPE_CHECKING:
    from typing import Any, Dict, Generator, Optional, Text, Tuple, Type


def _fields_name_from_attr(cls):
    # type: (Type) -> Generator[Tuple[Text, Optional[Text]]]
    if not attr.has(cls):
        raise TypeError("Class should be attrs based")
    for field in attr.fields(cls):
        if JsonInclude.THIS in field.metadata:
            yield field.name, None
        elif JsonInclude.NAME in field.metadata:
            yield field.name, camelcase_to_underscore(
                field.metadata.get(JsonInclude.NAME)
            )


def _types_and_fields_from_attr(cls):
    # type: (Type) -> Dict[Type, Any]
    types = {cls: tuple(_fields_name_from_attr(cls))}

    def traverse(cls):
        for field in attr.fields(cls):
            if field.type and attr.has(field.type):
                types[field.type] = tuple(_fields_name_from_attr(field.type))
                traverse(field.type)

    traverse(cls)
    return types


def _make_keys_as_underscore(dct):
    # type: (Dict[Text, Any]) -> Dict[Text, Any]
    params = {}
    for k, v in iteritems(dct):
        k = camelcase_to_underscore(k)
        params[k] = v
    return params


def _server_fields_name(fields):
    # type: (Tuple[Text, Optional[Text]]) -> Generator[Text]
    for loc_f, ser_f in fields:
        if ser_f:
            yield ser_f
            continue
        yield loc_f


def _cleaned_params(params, fields):
    # type: (Dict[Text, Any], Tuple[Text, Optional[Text]]) -> Dict[Text, Any]
    d = {}
    for key, val in iteritems(params):
        for loc_f, ser_f in fields:
            if key == loc_f or ser_f and key == ser_f:
                d[loc_f] = val
                break
    return d


def attr_from_json(content, cls):
    types_and_fields = _types_and_fields_from_attr(cls)

    def coincidence_find(params, coincidence, common_index=0):
        try:
            (kls, fields), _ = Counter(coincidence).most_common()[common_index]
            return kls(**_cleaned_params(params, fields))
        except TypeError:
            # in case of Region and AccessibilityRegion there could be situation when
            return coincidence_find(params, coincidence, common_index=common_index + 1)
        except IndexError:
            # Failed to convert any class. Use raw object instead
            return params

    def obj_came(obj):
        params = _make_keys_as_underscore(dict(obj))
        coincidence = defaultdict(int)
        for kls, fields in iteritems(types_and_fields):
            fields_to_compare = tuple(_server_fields_name(fields))
            if len(types_and_fields) == 1:
                return kls(**_cleaned_params(params, fields))
            if set(params.keys()) == set(fields_to_compare):
                return kls(**_cleaned_params(params, fields))

            for key in params.keys():
                if key in fields_to_compare:
                    coincidence[(kls, fields)] += 1
        return coincidence_find(params, coincidence)

    instance = json.loads(content, object_hook=obj_came)
    return instance


def camelcase_to_underscore(text):
    s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", text)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()


def underscore_to_camelcase(text):
    return re.sub(r"(?!^)_([a-zA-Z])", lambda m: m.group(1).upper(), text)


class JsonInclude(object):
    NON_NONE = "non_none"
    THIS = "include"
    NAME = "name"



from __future__ import absolute_import, unicode_literals

import typing
from typing import List, Optional, Text, Tuple, Union

from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.remote.webelement import WebElement
from six import string_types

from applitools.common import (
    EyesError,
    FailureReports,
    RectangleSize,
    TestFailedError,
    deprecated,
)
from applitools.common.selenium import Configuration

from .__version__ import __version__
from .command_executor import CommandExecutor
from .fluent.selenium_check_settings import SeleniumCheckSettings
from .fluent.target import Target
from .runner import ClassicRunner, EyesRunner, log_session_results_and_raise_exception
from .universal_sdk_types import (
    demarshal_locate_result,
    demarshal_match_result,
    demarshal_test_results,
    marshal_check_settings,
    marshal_configuration,
    marshal_locate_settings,
    marshal_ocr_extract_settings,
    marshal_ocr_search_settings,
    marshal_viewport_size,
    marshal_webdriver_ref,
)

if typing.TYPE_CHECKING:
    from applitools.common import MatchResult, Region, TestResults
    from applitools.common.utils.custom_types import FrameReference, ViewPort
    from applitools.core import (
        PositionProvider,
        TextRegionSettings,
        VisualLocatorSettings,
    )
    from applitools.core.extract_text import PATTERN_TEXT_REGIONS
    from applitools.core.locators import LOCATORS_TYPE
    from applitools.selenium import OCRRegion


class Eyes(object):
    def __init__(self, runner=None):
        # type: (Union[None, EyesRunner, Text]) -> None
        self.configure = Configuration()
        self._driver = None
        self._eyes_ref = None
        if runner is None:
            self._runner = ClassicRunner()
        elif isinstance(runner, string_types):
            self.configure.server_url = runner
            self._runner = ClassicRunner()
        else:
            self._runner = runner  # type: EyesRunner
        self._commands = self._runner._commands  # noqa

    def open(
        self,
        driver,  # type: WebDriver
        app_name=None,  # type: Optional[Text]
        test_name=None,  # type: Optional[Text]
        viewport_size=None,  # type: Optional[ViewPort]
    ):
        # type: (...) -> WebDriver
        if app_name is not None:
            self.configure.app_name = app_name
        if test_name is not None:
            self.configure.test_name = test_name
        if viewport_size is not None:
            self.configure.viewport_size = viewport_size
        if self.configure.app_name is None:
            raise ValueError("app_name should be set via configuration or an argument")
        if self.configure.test_name is None:
            raise ValueError("test_name should be set via configuration or an argument")

        if self.configure.is_disabled:
            pass
        else:
            self._driver = driver
            self._eyes_ref = self._commands.manager_open_eyes(
                self._runner._ref,  # noqa
                marshal_webdriver_ref(driver),
                self._marshaled_configuration(),
            )
        return driver

    @typing.overload
    def check(self, name, check_settings):
        # type: (Text, SeleniumCheckSettings) -> MatchResult
        """
        Takes a snapshot and matches it with the expected output.

        :param name: The name of the tag.
        :param check_settings: target which area of the window to check.
        """
        pass

    @typing.overload
    def check(self, check_settings):
        # type: (SeleniumCheckSettings) -> None
        """
        Takes a snapshot and matches it with the expected output.

        :param check_settings: target which area of the window to check.
        """
        pass

    def check(self, check_settings, name=None):
        # type: (SeleniumCheckSettings, Optional[Text]) -> Optional[MatchResult]
        if isinstance(name, SeleniumCheckSettings) or isinstance(
            check_settings, string_types
        ):
            check_settings, name = name, check_settings
        if check_settings is None:
            check_settings = Target.window()
        if name:
            check_settings = check_settings.with_name(name)

        if self.configure.is_disabled:
            return None
        if not self.is_open:
            self.abort()
            raise EyesError("you must call open() before checking")

        results = self._commands.eyes_check(
            self._eyes_ref,
            marshal_check_settings(_is_selenium_driver(self.driver), check_settings),
            self._marshaled_configuration(),
        )
        if results:
            results = demarshal_match_result(results)
            if (
                not results.as_expected
                and self.configure.failure_reports is FailureReports.IMMEDIATE
            ):
                raise TestFailedError(
                    "Mismatch found in '{}' of '{}'".format(
                        self.configure.test_name, self.configure.app_name
                    )
                )
            else:
                return results
        else:
            return None

    def locate(self, visual_locator_settings):
        # type: (VisualLocatorSettings) -> LOCATORS_TYPE
        results = self._commands.eyes_locate(
            self._eyes_ref,
            marshal_locate_settings(visual_locator_settings),
            self._marshaled_configuration(),
        )
        return demarshal_locate_result(results)

    def extract_text(self, *regions):
        # type: (*OCRRegion) -> List[Text]
        return self._commands.eyes_extract_text(
            self._eyes_ref,
            marshal_ocr_extract_settings(_is_selenium_driver(self.driver), regions),
            self._marshaled_configuration(),
        )

    def extract_text_regions(self, config):
        # type: (TextRegionSettings) -> PATTERN_TEXT_REGIONS
        return self._commands.eyes_extract_text_regions(
            self._eyes_ref,
            marshal_ocr_search_settings(config),
            self._marshaled_configuration(),
        )

    def close(self, raise_ex=True):
        # type: (bool) -> Optional[TestResults]
        """
        Ends the test.

        :param raise_ex: If true, an exception will be raised for failed/new tests.
        :return: The test results.
        """
        return self._close(raise_ex, True)

    def close_async(self):
        # type: () -> Optional[TestResults]
        return self._close(False, False)

    def abort(self):
        # type: () -> Optional[TestResults]
        """
        If a test is running, aborts it. Otherwise, does nothing.
        """
        return self._abort(True)

    def abort_async(self):
        return self._abort(False)

    @deprecated.attribute("use `abort()` instead")
    def abort_if_not_closed(self):
        return self.abort()

    @staticmethod
    def get_viewport_size(driver):
        # type: (WebDriver) -> RectangleSize
        cmd = CommandExecutor.get_instance(EyesRunner.BASE_AGENT_ID, __version__)
        result = cmd.core_get_viewport_size(marshal_webdriver_ref(driver))
        return RectangleSize.from_(result)

    @staticmethod
    def set_viewport_size(driver, viewport_size):
        # type: (WebDriver, ViewPort) -> None
        cmd = CommandExecutor.get_instance(EyesRunner.BASE_AGENT_ID, __version__)
        cmd.core_set_viewport_size(
            marshal_webdriver_ref(driver), marshal_viewport_size(viewport_size)
        )

    @property
    def configuration(self):
        return self.configure

    @configuration.setter
    def configuration(self, value):
        self.configure = value

    def get_configuration(self):
        # type:() -> Configuration
        return self.configure.clone()

    def set_configuration(self, configuration):
        # type:(Configuration) -> None
        self.configure = configuration.clone()

    @property
    def is_open(self):
        # type: () -> bool
        return self._eyes_ref is not None

    @property
    def server_connector(self):
        raise NotImplementedError

    @server_connector.setter
    def server_connector(self, server_connector):
        raise NotImplementedError

    @property
    def base_agent_id(self):
        # type: () -> Text
        """
        Must return version of SDK. (e.g. selenium, visualgrid) in next format:
            "eyes.{package}.python/{lib_version}"
        """
        return "{}/{}".format(self._runner.BASE_AGENT_ID, __version__)

    @property
    def full_agent_id(self):
        # type: () -> Text
        """
        Gets the agent id, which identifies the current library using the SDK.

        """
        if self.configure.agent_id is None:
            return self.base_agent_id
        else:
            return "{} [{}]".format(self.configure.agent_id, self.base_agent_id)

    @property
    def is_cut_provider_explicitly_set(self):
        # type: () -> bool
        """
        Gets is cut provider explicitly set.
        """
        raise self.configure.cut_provider is not None

    @property
    def driver(self):
        # type: () -> WebDriver
        return self._driver

    def check_window(self, tag=None, match_timeout=-1, fully=None):
        # type: (Optional[Text], int, Optional[bool]) -> MatchResult
        """
        Takes a snapshot of the application under test and matches it with the expected
         output.

        :param tag: An optional tag to be associated with the snapshot.
        :param match_timeout:  The amount of time to retry matching (milliseconds)
        :param fully: Defines that the screenshot will contain the entire window.
        """
        return self.check(tag, Target.window().timeout(match_timeout).fully(fully))

    def check_frame(self, frame_reference, tag=None, match_timeout=-1):
        # type: (FrameReference, Optional[Text], int) -> MatchResult
        """
        Check frame.

        :param frame_reference: The name or id of the frame to check. (The same
                name/id as would be used in a call to driver.switch_to.frame()).
        :param tag: An optional tag to be associated with the match.
        :param match_timeout: The amount of time to retry matching. (Milliseconds)
        """
        return self.check(
            tag, Target.frame(frame_reference).fully().timeout(match_timeout)
        )

    def check_region(
        self,
        region,  # type: Union[Region, Text, List, Tuple, WebElement]
        tag=None,  # type: Optional[Text]
        match_timeout=-1,  # type: int
        stitch_content=False,  # type: bool
    ):
        # type: (...) -> MatchResult
        """
        Takes a snapshot of the given region from the browser using the web driver
        and matches it with the expected output. If the current context is a frame,
        the region is offsetted relative to the frame.

        :param region: The region which will be visually validated. The coordinates are
                       relative to the viewport of the current frame.
        :param tag: Description of the visual validation checkpoint.
        :param match_timeout: Timeout for the visual validation checkpoint
                              (milliseconds).
        :param stitch_content: If `True`, stitch the internal content of the region
        """
        return self.check(
            tag,
            Target.region(region).timeout(match_timeout).fully(stitch_content),
        )

    def check_element(
        self,
        element,  # type: Union[Text,List,Tuple,WebElement]
        tag=None,  # type: Optional[Text]
        match_timeout=-1,  # type: int
    ):
        # type: (...) -> MatchResult
        """
        Takes a snapshot of the given region from the browser using the web driver
        and matches it with the expected output. If the current context is a frame,
        the region is offsetted relative to the frame.

        :param element: The element to check.
        :param tag: Description of the visual validation checkpoint.
        :param match_timeout: Timeout for the visual validation checkpoint
                              (milliseconds).
        """
        return self.check(
            tag,
            Target.region(element).timeout(match_timeout).fully(),
        )

    def check_region_in_frame(
        self,
        frame_reference,  # type: FrameReference
        region,  # type: Union[Region, Text, List, Tuple, WebElement]
        tag=None,  # type: Optional[Text]
        match_timeout=-1,  # type: int
        stitch_content=False,  # type: bool
    ):
        # type: (...) -> MatchResult
        """
        Checks a region within a frame, and returns to the current frame.

        :param frame_reference: A reference to the frame in which the region
                                should be checked.
        :param region: Specifying the region to check inside the frame.
        :param tag: Description of the visual validation checkpoint.
        :param match_timeout: Timeout for the visual validation checkpoint
                              (milliseconds).
        :param stitch_content: If `True`, stitch the internal content of the region
        """
        return self.check(
            tag,
            Target.region(region)
            .frame(frame_reference)
            .stitch_content(stitch_content)
            .timeout(match_timeout),
        )

    # Impossible to implement via universal sdk
    @property
    def should_stitch_content(self):
        # type: () -> bool
        raise NotImplementedError

    @property
    def original_fc(self):
        """Gets original frame chain

        Before check() call we save original frame chain

        Returns:
            Frame chain saved before check() call
        """
        raise NotImplementedError

    @property
    def device_pixel_ratio(self):
        # type: () -> int
        """
        Gets device pixel ratio.

        :return The device pixel ratio, or if the DPR is not known yet or if it wasn't
        possible to extract it.
        """
        raise NotImplementedError

    @property
    def debug_screenshots_provider(self):
        raise NotImplementedError

    @property
    def position_provider(self):
        """
        Sets position provider.
        """
        raise NotImplementedError

    @property
    def current_frame_position_provider(self):
        # type: () -> Optional[PositionProvider]
        raise NotImplementedError

    def add_mouse_trigger_by_element(self, action, element):
        # type: (Text, WebElement) -> None
        """
        Adds a mouse trigger.

        :param action: Mouse action (click, double click etc.)
        :param element: The element on which the action was performed.
        """
        raise NotImplementedError

    def add_text_trigger_by_element(self, element, text):
        # type: (WebElement, Text) -> None
        """
        Adds a text trigger.

        :param element: The element to which the text was sent.
        :param text: The trigger's text.
        """
        raise NotImplementedError

    @property
    def agent_setup(self):
        # Saved for backward compatibility
        return None

    def _marshaled_configuration(self):
        return marshal_configuration(_is_selenium_driver(self.driver), self.configure)

    def _close(self, raise_ex, wait_result):
        # type: (bool, bool) -> Optional[TestResults]
        if self.configure.is_disabled:
            return None
        if not self.is_open:
            raise EyesError("Eyes not open")
        results = self._commands.eyes_close_eyes(self._eyes_ref, wait_result)
        self._eyes_ref = None
        self._driver = None
        if wait_result:
            results = demarshal_test_results(results, self.configure)
            for r in results:
                log_session_results_and_raise_exception(raise_ex, r)
            return results[0]  # Original interface returns just one result
        else:
            return None

    def _abort(self, wait_result):
        # type: (bool) -> Optional[List[TestResults]]
        if self.configure.is_disabled:
            return None
        elif self.is_open:
            results = self._commands.eyes_abort_eyes(self._eyes_ref, wait_result)
            self._eyes_ref = None
            self._driver = None
            if wait_result:
                return demarshal_test_results(results, self.configure)
            else:
                return None

    def __getattr__(self, item):
        return getattr(self.configure, item)

    def __setattr__(self, key, value):
        if "configure" in vars(self) and (
            key in vars(self.configure)
            or key in ("match_level", "ignore_displacements")
        ):
            return setattr(self.configure, key, value)
        else:
            return super(Eyes, self).__setattr__(key, value)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            self._abort(self._runner.AUTO_CLOSE_MODE_SYNC)
        else:
            self._close(True, self._runner.AUTO_CLOSE_MODE_SYNC)


def _is_selenium_driver(driver):
    # type: (WebDriver) -> bool
    try:
        from appium.webdriver.webdriver import WebDriver as AppiumWebDriver

        # appium.WebDriver is based on selenium.WebDriver so there is no sense to check
        # for relationship with selenium.WebDriver. Have to eliminate appium instead
        return not isinstance(driver, AppiumWebDriver)
    except ImportError:
        return True



import warnings
from enum import Enum

from applitools.common.utils.general_utils import DynamicEnumGetter


class BrowserType(Enum):
    CHROME = "chrome"
    CHROME_ONE_VERSION_BACK = "chrome-one-version-back"
    CHROME_TWO_VERSIONS_BACK = "chrome-two-versions-back"
    FIREFOX = "firefox"
    FIREFOX_ONE_VERSION_BACK = "firefox-one-version-back"
    FIREFOX_TWO_VERSIONS_BACK = "firefox-two-versions-back"
    SAFARI = "safari"
    SAFARI_ONE_VERSION_BACK = "safari-one-version-back"
    SAFARI_TWO_VERSIONS_BACK = "safari-two-versions-back"
    SAFARI_EARLY_ACCESS = "safari-earlyaccess"
    IE_10 = "ie10"
    IE_11 = "ie11"
    EDGE_LEGACY = "edgelegacy"
    EDGE_CHROMIUM = "edgechromium"
    EDGE_CHROMIUM_ONE_VERSION_BACK = "edgechromium-one-version-back"
    EDGE_CHROMIUM_TWO_VERSIONS_BACK = "edgechromium-two-versions-back"

    @DynamicEnumGetter
    def EDGE(self):
        # type: () -> BrowserType
        warnings.warn(
            "The `EDGE` option that is being used in your browsers configuration"
            " will soon be deprecated. Please change it to either `EDGE_LEGACY`"
            " for the legacy version or to `EDGE_CHROMIUM` for the"
            " new Chromium-based version.",
            DeprecationWarning,
            stacklevel=3,
        )
        return self.EDGE_LEGACY


class StitchMode(Enum):
    """
    The type of methods for stitching full-page screenshots.
    """

    Scroll = "Scroll"
    CSS = "CSS"



import typing

import attr

from applitools.common import FloatingBounds
from applitools.common.accessibility import AccessibilityRegionType
from applitools.common.geometry import AccessibilityRegion, Rectangle, Region

if typing.TYPE_CHECKING:
    from typing import List, Optional, Union

    from applitools.common.utils.custom_types import CodedRegionPadding

__all__ = (
    "GetFloatingRegion",
    "GetRegion",
    "RegionByRectangle",
    "FloatingRegionByRectangle",
    "GetAccessibilityRegion",
    "AccessibilityRegionByRectangle",
)


class GetRegion(object):
    @property
    def padding(self):
        # type: () -> Optional[CodedRegionPadding]
        return getattr(self, "_padding", None)


class GetFloatingRegion(GetRegion):
    pass


class GetAccessibilityRegion(GetRegion):
    pass


@attr.s
class RegionByRectangle(GetRegion):
    _region = attr.ib()  # type: Union[Region, Rectangle]


@attr.s
class FloatingRegionByRectangle(GetFloatingRegion):
    _rect = attr.ib()  # type: Union[Region, Rectangle]
    _bounds = attr.ib()  # type: FloatingBounds

    @property
    def floating_bounds(self):
        return self._bounds


@attr.s
class AccessibilityRegionByRectangle(GetAccessibilityRegion):
    _rect = attr.ib()  # type: Union[Region, Rectangle, AccessibilityRegion]
    _type = attr.ib(default=None)  # type: Optional[AccessibilityRegionType]

    @property
    def accessibility_type(self):
        # type: () -> AccessibilityRegionType
        if self._type:
            return self._type
        return self._rect.type



from . import geometry  # noqa
from . import (
    accessibility,
    config,
    errors,
    logger,
    match,
    selenium,
    server,
    test_results,
    ultrafastgrid,
)
from .accessibility import (  # noqa
    AccessibilityGuidelinesVersion,
    AccessibilityLevel,
    AccessibilityRegionType,
    AccessibilitySettings,
)
from .config import BatchInfo, Configuration, ProxySettings  # noqa
from .errors import (  # noqa
    DiffsFoundError,
    EyesError,
    NewTestError,
    OutOfBoundsError,
    TestFailedError,
)
from .geometry import (  # noqa
    AccessibilityRegion,
    CoordinatesType,
    Point,
    RectangleSize,
    Region,
    SubregionForStitching,
)
from .logger import FileLogger, StdoutLogger  # noqa
from .match import (  # noqa
    ExactMatchSettings,
    FloatingBounds,
    FloatingMatchSettings,
    ImageMatchSettings,
    MatchLevel,
    MatchResult,
)
from .selenium import StitchMode  # noqa
from .server import FailureReports, SessionType  # noqa
from .test_results import TestResultContainer, TestResults, TestResultsSummary  # noqa
from .ultrafastgrid.config import (  # noqa
    DeviceName,
    IosDeviceName,
    IosVersion,
    ScreenOrientation,
    VisualGridOption,
)
from .ultrafastgrid.render_browser_info import (
    ChromeEmulationInfo,
    DesktopBrowserInfo,
    IosDeviceInfo,
    RenderBrowserInfo,
)

__all__ = (
    logger.__all__  # noqa
    + errors.__all__  # noqa
    + geometry.__all__  # noqa
    + match.__all__  # noqa
    + test_results.__all__  # noqa
    + server.__all__  # noqa
    + (
        "logger",
        "StitchMode",
        "ChromeEmulationInfo",
        "DesktopBrowserInfo",
        "DeviceName",
        "ScreenOrientation",
        "AccessibilityRegionType",
        "AccessibilitySettings",
        "AccessibilityLevel",
        "AccessibilityGuidelinesVersion",
    )  # noqa
)
